import os
import sys
import io
import random
from threading import Thread
from queue import Queue
import time


import numpy as np
import aubio
import pygame as pg
pygame = pg

from parrotjoy.metronome import Bpm, BpmCounter, BpmLight, BpmLine
from parrotjoy.audiorecord import PITCHES, ONSETS, AUDIOS
from parrotjoy.draw_sound import draw_wave
from parrotjoy.camera_cv import VideoThread
from parrotjoy.tracks import Track
from parrotjoy.resources import gfx


class Joy:
    """ For keeping track of if joy buttons were pressed down.
    """
    def __init__(self):
        self.joys = {}
        self.joys[0] = {}
        self.joys[0]['buttons'] = {}
        for x in range(20):
            self.joys[0]['buttons'][x] = 0

    def events(self, events):
        for e in events:
            if e.type == pg.JOYBUTTONDOWN:
                self.joys[0]['buttons'][e.button] = 1
            if e.type == pg.JOYBUTTONUP:
                self.joys[0]['buttons'][e.button] = 0


JOY_SELECT = 8
JOY_START = 9
JOY_1 = 0
JOY_2 = 1
JOY_3 = 2
JOY_4 = 3
JOY_R1 = 5
JOY_R2 = 7
JOY_L1 = 4
JOY_L2 = 6

TRIM_AMOUNT = 0.1


class TrackRecorder:
    """ For recording, and playing back different tracks.
    """
    def __init__(self, audio_thread):


        self.audio_thread = audio_thread


        # busy loop until we get some sound coming in.
        # This syncs the 'bpm' timer to the sound input.
        found = False
        while 1:
            if not self.audio_thread.audio_queue.empty():
                self.bpm = Bpm()
                break
            time.sleep(0.0001)

        self.tracks = [Track(self.bpm) for x in range(4)]
        self.track_idx = 0
        self.track = self.tracks[self.track_idx]

        self.joys = Joy()
        self.joy_buttons = self.joys.joys[0]['buttons']
        self.playing = False
        """ True if we should play the sounds.
        """


    def play(self):
        print("TrackRecorder: play")
        if self.playing:
            first_two_tracks = self.tracks[:2]
            for track in first_two_tracks:
                print(track.sounds)
                for sound in track.sounds:
                    channel = sound.play()
                    # channel.set_endevent(pg.USEREVENT+2)
                    print(self.bpm.time_since_start(0.0), sound.get_length(), self.bpm._elapsed_time, self.bpm.get_time_for_loop())

    def start(self):
        """ Tell it to start playing the sounds.
        """
        self.stop()
        self.playing = True
        self.bpm.start_at_beginning()


    def stop(self):
        """ Tell it to STOP playing the sounds.
        """
        self.playing = False
        for track in self.tracks:
            for sound in track.sounds:
                sound.stop()

    def event_joy(self, e):
        track_buttons = [JOY_1, JOY_2, JOY_3, JOY_4]
        if e.type == pg.JOYBUTTONDOWN and e.button in track_buttons:
            track_idx = track_buttons.index(e.button)
            track = self.tracks[track_idx]
            self.track_idx = track_idx
            if self.joy_buttons[JOY_R1]:
                track.start_new_next()
            elif self.joy_buttons[JOY_R2]:
                track.add_to_next()
            elif self.joy_buttons[JOY_L1]:
                track.erase()
                if track_idx == 0:
                    self.bpm.init(70, 60/70.0)
            elif track.recording:
                # r1, r2, l1 not pressed, we are recording.
                if track.recording == 2:
                    track.finish()
            else:
                track.play()


    def trim_audio(self, start=0, end=0):
        track = self.tracks[self.track_idx]
        track.trim_audio(start=start, end=end)
        track.stop()
        track.play()


    def events(self, events):

        # print('bpm updating...')
        self.bpm.update()

        # update the tracks because sometimes there is no audio
        for track in self.tracks:
            track.update(None)

        self.joys.events(events)

        for e in events:
            # print("events", e)
            if e.type == pg.QUIT:
                running = False
            elif e.type == PITCHES:
                for pitch in e.pitches:
                    pass
            elif e.type == ONSETS:
                for onset in e.onsets:
                    pass
            elif e.type == AUDIOS:
                for audio in e.audios:
                    for track in self.tracks:
                        track.update(audio)

            if e.type == pg.KEYDOWN:
                if e.key == pg.K_z:
                    self.audio_thread.audio_going = False
                if e.key == pg.K_s:
                    self.track.start_new_next()
                if e.key == pygame.K_s and pygame.key.get_mods() & pygame.KMOD_SHIFT:
                    self.track.add_to_next()
                if e.key == pg.K_p:
                    self.track.play()
                if e.key == pg.K_f:
                    self.track.finish()
                    # if track.sounds:
                    #     for sound in track.sounds[-5]:
                    #         sound.play()
                if e.key == pg.K_w:
                    self.trim_audio(start=-TRIM_AMOUNT)
                if e.key == pg.K_e:
                    self.trim_audio(start=TRIM_AMOUNT)
                if e.key == pg.K_r:
                    self.trim_audio(end=TRIM_AMOUNT)
                if e.key == pg.K_t:
                    self.trim_audio(end=-TRIM_AMOUNT)

            self.event_joy(e)

            if e.type == pg.JOYAXISMOTION:
                if e.axis == 0 and e.value < -0.01:
                    # lpaddle left
                    self.trim_audio(start=-TRIM_AMOUNT)
                elif e.axis == 0 and e.value >= 0.01:
                    # lpaddle right
                    self.trim_audio(start=TRIM_AMOUNT)
                elif e.axis == 2 and e.value < -0.01:
                    #rpaddle left
                    self.trim_audio(end=TRIM_AMOUNT)
                elif e.axis == 2 and e.value >= 0.01:
                    #rpaddle right
                    self.trim_audio(end=-TRIM_AMOUNT)


            if e.type == pg.JOYBUTTONDOWN and e.button == JOY_SELECT:
                self.stop()
            if e.type == pg.JOYBUTTONDOWN and e.button == JOY_START:
                self.start()
                pass
                # self.track.add_to()
                # self.track.finish()

            if e.type == pg.JOYBUTTONUP:
                pass
                # self.joys[0]['buttons'][e.button] = 0

        # if the first track was just finished, we see how big it is.
        if ((self.tracks[0].finished or self.tracks[0].trimmed) or
            (self.tracks[1].finished or self.tracks[1].trimmed)):
            # get the longest of the two looping tracks.
            sound_length1 = 0
            sound_length2 = 0
            if self.tracks[0].sounds:
                sound_length1 = self.tracks[0].sounds[0].get_length()
            if self.tracks[1].sounds:
                sound_length2 = self.tracks[1].sounds[0].get_length()

            sound_length = max(sound_length1, sound_length2)

            space_between_beats = sound_length / 4.
            bpm = 60 / space_between_beats
            self.bpm.init(bpm, space_between_beats)
            self.start()


class RecordingWave(pg.sprite.DirtySprite):
    """ Shows a
    """
    def __init__(self, track, track_idx):
        pg.sprite.DirtySprite.__init__(self)
        self.track = track
        self.track_idx = track_idx

        # how many pixels wide for how many seconds?
        self.full_width_seconds = 5
        self.wave_width = 320

        self.image_height = 130
        self.background_color = (0, 0, 0)
        self.image = pg.Surface((320, self.image_height)).convert()
        self.image.fill(self.background_color)
        self.rect = self.image.get_rect()
        self.rect.x = 10
        self.rect.y = 70 + (self.track_idx * self.image_height)
        self.state = None

    def draw_wave(self):

        # How big should the image be?
        #    Depends on how much sound is recorded.
        sound_length = self.track.sounds[0].get_length()
        image_width = self.wave_width / self.full_width_seconds * sound_length
        self.image = pg.Surface((image_width, self.image_height)).convert()
        self.rect = self.image.get_rect()
        self.rect.x = 10
        self.rect.y = 70 + (self.track_idx * self.image_height)


        float32_data = np.frombuffer(self.track.audios[0], dtype=np.float32)
        int16_data = ((float32_data / 1.414) * pow(2, 16) // 2).astype(np.int16)
        samples = int16_data

        draw_wave(
            self.image,
            samples,
            background_color=self.background_color,
            wave_color=(255, 255, 255))


    def update(self):
        """
        """
        # print("RecordingWave: self.track.recording, self.state, self.track.recording, self.track.add_to_mode")
        # print(self.track.recording, self.state, self.track.recording, self.track.add_to_mode)
        if self.track.recording == 2 and (not self.track.add_to_mode) and self.state != 'recording':
            print('WAVE: recording')
            self.image.fill((0, 0, 0))
            self.dirty = 1
            self.state = 'recording'
        elif self.track.add_to_mode and self.state != 'add_to_mode':
            print('WAVE: add_to_mode')
            self.state = 'add_to_mode'
        elif (
            (not (self.track.recording or self.track.add_to_mode)) and
            self.state in ['add_to_mode', 'recording']
        ):
            print('WAVE: update wave')
            if self.track.sounds and self.track.sounds[0]:
                self.draw_wave()
            self.dirty = 1
            self.state = None
        elif self.track.erased:
            print('WAVE: erased')
            self.image.fill((0, 0, 0))
            self.dirty = 1
        elif self.track.trimmed:
            print('WAVE: trimmed')
            if self.track.sounds and self.track.sounds[0]:
                self.draw_wave()
            self.dirty = 1
            self.state = None


class RecordingLight(pg.sprite.DirtySprite):
    """ Shows a red 'light' representing each beat.
    """
    def __init__(self, track, track_idx):
        pg.sprite.DirtySprite.__init__(self)
        self.track = track
        self.track_idx = track_idx

        self.image = pg.Surface((20, 20))
        self.image.fill((25, 25, 25))
        self.rect = self.image.get_rect().copy()
        self.rect.x = 50 + (track_idx * 30)
        self.rect.y = 30
        self.state = None

    def update(self):

        if self.track.recording == 1 and (not self.track.add_to_mode) and self.state != 'preparing':
            self.image.fill((0, 0, 255))
            self.dirty = 1
            self.state = 'preparing'
        elif self.track.recording == 2 and (not self.track.add_to_mode) and self.state != 'recording':
            self.image.fill((0, 255, 255))
            self.dirty = 1
            self.state = 'recording'
        elif self.track.add_to_mode and self.state != 'add_to_mode':
            self.image.fill((0, 255, 0))
            self.dirty = 1
            self.state = 'add_to_mode'
        elif (
            (not (self.track.recording or self.track.add_to_mode)) and
            self.state in ['add_to_mode', 'recording']
        ):
            self.image.fill((25, 25, 25))
            self.dirty = 1
            self.state = None
        # print (self.state, self.track.recording, self.track.add_to_mode)


class Parrot(pg.sprite.DirtySprite):
    """
    """
    def __init__(self):
        pg.sprite.DirtySprite.__init__(self)
        self.image = gfx('parrotjoy-logo.png')
        self.rect = self.image.get_rect().copy()
        self.rect.x = pg.display.get_surface().get_width() - self.rect.width
        self.rect.y = 20
        # self.dirty = 1


class Gif:
    def __init__(self):
        self.path = '/tmp/'
        self.start_saving = False
        self.finished_saving = False
        self.surfs = []
        self.fps = 30
    def finish(self):
        print("saving gifs")
        image_paths = []
        for frame_idx, surf in enumerate(self.surfs):
            image_path = f'{self.path}/bla_%05d.png' % frame_idx
            image_paths.append(image_path)
            pygame.image.save(surf, image_path)

        convertpath = 'convert'
        cmd = [
            convertpath,
            "-delay", f"{1000 // self.fps},1000",
            "-size", f'{self.surfs[0].get_width()}x{self.surfs[0].get_height()}']
        cmd += image_paths
        cmd += [f"{self.path}/anim.gif"]
        print(cmd)
        import subprocess
        subprocess.call(cmd)

        for image_path in image_paths:
            os.remove(image_path)

        self.image_path = []
        self.finished_saving = False

    def update(self, events, screen):
        for e in events:
            if e.type == pg.KEYDOWN:
                if e.key == pg.K_a and not self.start_saving:
                    self.start_saving = True
                    self.finished_saving = False
                    print("recording surfs, press a")
                elif e.key == pg.K_a and self.start_saving:
                    self.start_saving = False
                    self.finished_saving = True
        if self.finished_saving:
            self.finish()
        if self.start_saving:
            self.surfs.append(screen.copy())


# CAMERA_RES = list(map(int, [640//1.3, 480//1.3]))
CAMERA_RES = [480, 360]

class Looper:
    def __init__(self, app):
        self.active = True

        self.screen = app.screen
        self._app = app

        # We loop quickly so timing can be more accurate with the sounds.
        self.fps = 240

        #TODO: FIXME: this camera id should be selectable.
        camera_id = 0
        self.video_thread = VideoThread(camera_id, CAMERA_RES[0], CAMERA_RES[1])
        self.video_thread.daemon = True
        self.video_thread.start()



        track_recorder = TrackRecorder(app.audio_thread)
        self.track_recorder = track_recorder

        recording_lights = [
            RecordingLight(track, idx)
            for idx, track in enumerate(track_recorder.tracks)
        ]
        self.recording_lights = recording_lights

        recording_waves = [
            RecordingWave(track, idx)
            for idx, track in enumerate(track_recorder.tracks)
        ]
        self.recording_waves = recording_waves


        bpm_counter = BpmCounter(track_recorder.bpm)
        self.bpm_counter = bpm_counter

        bpm_light = BpmLight(track_recorder.bpm)
        self.bpm_light = bpm_light

        bpm_line = BpmLine(track_recorder.bpm)
        self.bpm_line = bpm_line


        self.parrot = Parrot()

        pg.display.set_caption('press space 4 times to adjust BPM timing')
        screen = app.screen

        background = pg.Surface(screen.get_size())
        background = background.convert()
        background.fill((0, 0, 0))
        self.background = background

        sprite_list = [bpm_counter, bpm_light, bpm_line] + recording_lights + recording_waves + [self.parrot]
        allsprites = pg.sprite.LayeredDirty(
            sprite_list,
            _time_threshold=1000/10.0
        )
        allsprites.clear(screen, background)
        self.allsprites = allsprites


    def redraw(self):
        screen = self.screen
        bpm_counter = self.bpm_counter
        track_recorder = self.track_recorder
        allsprites = self.allsprites
        background = self.background

        rects = []
        rects.append(screen.blit(background, (0, 0)))
        allsprites.clear(screen, background)
        for spr in allsprites:
            spr.dirty = 1
        allsprites.draw(screen)
        pg.display.update(rects)

    def events(self, events):
        rects = []

        screen = self.screen
        bpm_counter = self.bpm_counter
        track_recorder = self.track_recorder
        allsprites = self.allsprites
        video_thread = self.video_thread
        background = self.background

        for e in events:
            if e.type == pg.QUIT or e.type == pg.KEYDOWN and e.key in [pg.K_ESCAPE, pg.K_q]:
                going = False

            if e.type == pg.KEYDOWN and e.key in [pg.K_c]:
                self.redraw()

                # allsprites.clear(screen, background)
                video_thread.stop()
                # video_thread = VideoThread(1, 1920, 1080)
                video_thread = VideoThread(0, CAMERA_RES[0], CAMERA_RES[1])
                video_thread.daemon = True
                video_thread.start()
            if e.type == pg.KEYDOWN and e.key in [pg.K_v]:
                screen.blit(background, (0, 0))
                allsprites.clear(screen, background)
                video_thread.stop()
                # video_thread = VideoThread(1, 1920, 1080)
                video_thread = VideoThread(0, CAMERA_RES[0], CAMERA_RES[1])
                video_thread.daemon = True
                video_thread.start()


        track_recorder.events(events)


        if track_recorder.bpm.started_beat and track_recorder.bpm.first_beat:
            track_recorder.play()

        bpm_counter.events(events)

        self.rects = rects



    def update(self, elapsed_time):
        """ return True to let other scenes update. False to only us update.
        """
        pass

    def render(self):
        """ return rects.

            If scene.propagate_render is True, the render will
                continue to be propagated.
        """
        rects = self.rects
        screen = self.screen
        bpm_counter = self.bpm_counter
        track_recorder = self.track_recorder
        allsprites = self.allsprites
        video_thread = self.video_thread
        bpm_line = self.bpm_line


        # the line changes color depending on when we are recording.
        for track in track_recorder.tracks:
            if track.recording == 1:
                bpm_line.color = bpm_line.preparing_color
                break
            elif track.recording == 2:
                bpm_line.color = bpm_line.recording_color
                break
            else:
                bpm_line.color = bpm_line.not_recording_color


        if video_thread.width != 1920:
            allsprites.update()
            rects.extend(allsprites.draw(screen))


        while not video_thread.queue.empty():
            surface = video_thread.queue.get()
            if surface:
                if video_thread.width == CAMERA_RES[0]:
                    video_x = screen.get_width() - CAMERA_RES[0]
                    video_y = (screen.get_height() - CAMERA_RES[1]) - 25
                else:
                    video_x = 0
                    video_y = 0
                video_rect = screen.blit(surface, (video_x, video_y))
                rects.append(video_rect)
        return rects



"""
Show a light for every beat.
Show the BPM as text.
Use 4/4 timing. Four beats to a bar.

Press space until the BPM gets to what you want.
"""
import time
import os
import pygame as pg
from typing import List


def average_bpm(times: List[float]) -> float:
    """For the list of times(seconds since epoch) return
    the average beats per minute.
    """
    spaces = []
    previous = times[0]
    for t in times[1:]:
        spaces.append(t - previous)
        previous = t
    avg_space = sum(spaces) / len(spaces)
    return 60.0 / avg_space


class Bpm:
    """Beats Per Minute.

    By press()ing this, the bpm counter will change.
    """

    def __init__(self):

        self.init()

    def init(self, bpm=70, space_between_beats=60 / 70):
        # now = time.time()
        # start the elapsed time after a chunk of sound has arrived.
        now = time.time() - (1.0 / (44100 / 512.0))
        self.space_between_beats = space_between_beats
        self.last_press: float = now
        """ The time since epoch of the last press.
        """
        self.bpm: int = bpm
        self.bpm_average: int = bpm
        """ The average bpm from the last 4 presses.
        """

        self.times: List[float] = []

        self._elapsed_time = 0.0
        self._last_update = now
        # self._elapsed_time = 1.0 / (44100 / 512.)
        self._last_closeness = 1.0
        self._last_start = 1

        self.on_beat: int | float = 0
        """ The time since the epoch when the last beat happened.
        """

        self.beat_num = 0
        """ accumulative counter of beats.
        """

        self.finished_beat = 0
        """ True for one tick, when 0.1 seconds has passed since the beat.
        """

        self.first_beat = 0
        """ True if we are the first beat on a bar (out of 4).
        """

        self.started_beat = 0
        """ This is true only on the tick
        """

    def start_at_beginning(self):
        self.init(self.bpm_average, self.space_between_beats)

    def press(self):
        """For when someone is trying to update the timer."""
        press_time = time.time()
        self.space_between_beats = press_time - self.last_press
        self.last_press = press_time
        self.times.append(press_time)
        self.bpm = 60.0 / self.space_between_beats

        if len(self.times) > 4:
            self.bpm_average = average_bpm(self.times[-4:])
            self.times = self.times[-4:]
        else:
            self.bpm_average = self.bpm

    def get_space_between(self):
        return 60.0 / self.bpm_average

    def get_time_for_loop(self):
        return 4 * self.get_space_between()

    def time_since_start(self, mod=0):
        return (self._elapsed_time + mod) % (self.get_time_for_loop())

    def update(self):
        # print('bpm: update')
        the_time = time.time()
        self._elapsed_time += the_time - self._last_update
        self._last_update = the_time

        # if _elapsed_time 'close' to bpm show light.
        since_last_beat = the_time - self.on_beat

        self.finished_beat = self.on_beat and (since_last_beat > 0.1)
        if self.finished_beat:
            self.on_beat = 0

        closeness = self._elapsed_time % self.get_space_between()
        if closeness < self._last_closeness:
            self.on_beat = the_time
            self.finished_beat = 0
            self.dirty = 1
            self.beat_num += 1
            self.started_beat = 1
            # self.first_beat = not (self.beat_num % 4)
            time_since_start = self.time_since_start()
            if time_since_start < self._last_start:
                self.first_beat = True
            else:
                self.first_beat = False
            self._last_start = time_since_start
            # print("-----1--- self.first_beat", self.first_beat, self.started_beat, closeness, self.beat_num, self.time_since_start())
        else:
            # print("self.started_beat = 0")
            self.started_beat = 0

        self._last_closeness = closeness


class BpmCounter(pg.sprite.DirtySprite):
    """For showing a text representing what the Beats Per Minute is."""

    def __init__(self, bpm):
        pg.sprite.DirtySprite.__init__(self)
        self.bpm = bpm
        self.last_bpm = self.bpm.bpm_average

        # self.font = pg.font.SysFont("Arial", 24)
        self.font = pg.font.Font(None, 24)
        self.render_image()

    def render_image(self):
        self.image = self.font.render(
            "%s" % int(self.bpm.bpm_average), True, (255, 255, 255)
        )
        self.rect = self.image.get_rect()

    def press(self):
        self.bpm.press()
        self.image = self.font.render(
            "%s" % int(self.bpm.bpm_average), True, (255, 255, 255)
        )
        self.rect = self.image.get_rect()
        self.dirty = 1

    def update(self):

        if self.last_bpm != self.bpm.bpm_average:
            self.render_image()
            self.dirty = 1
        self.last_bpm = self.bpm.bpm_average

    def events(self, events):
        for e in events:
            if e.type == pg.KEYDOWN and e.key == pg.K_SPACE:
                self.press()


class BpmLight(pg.sprite.DirtySprite):
    """Shows a red 'light' representing each beat."""

    def __init__(self, bpm):
        pg.sprite.DirtySprite.__init__(self)
        self.bpm = bpm

        self.image = pg.Surface((20, 20))
        self.image.fill((255, 0, 0))
        self.rect = self.image.get_rect()
        self.rect.x = 50
        self.rect.y = 5
        self.last_bpm = self.bpm.bpm_average

    def press(self):
        self.image.fill((255, 0, 0))
        self.elapsed_time = 0.0

    def update(self):

        if self.bpm.finished_beat:
            self.image.fill((0, 0, 0))
            self.dirty = 1
        if self.bpm.started_beat:
            self.image.fill((255, 0, 0))
            self.dirty = 1

        if self.last_bpm != self.bpm.bpm_average:
            self.dirty = 1
        self.last_bpm = self.bpm.bpm_average


class BpmLine(pg.sprite.DirtySprite):
    """Shows a line of where we are up to in a loop."""

    def __init__(self, bpm):
        self._layer = 1
        pg.sprite.DirtySprite.__init__(self)
        self.bpm = bpm

        self.preparing_color = (255, 255, 0)
        self.recording_color = (255, 0, 0)
        self.not_recording_color = (255, 255, 255)

        self.color = self.not_recording_color
        self.last_color = self.color

        self.render_image()
        self.last_bpm = self.bpm.bpm_average

    def render_image(self):
        # how wide is it for the whole thingo in pixels?
        width_pixels = 320
        width_in_seconds = 5
        print("self.bpm.get_time_for_loop()", self.bpm.get_time_for_loop())
        self.loop_width = (
            width_pixels / width_in_seconds
        ) * self.bpm.get_time_for_loop()

        # self.image_height = 130
        num_tracks_high = 4
        # self.image_height = 70 + (130 * num_tracks_high)
        self.image_height = 70 + (130 * 1)

        self.image = pg.Surface((1, self.image_height)).convert()
        self.image.fill((0, 0, 0))
        pg.draw.line(self.image, self.color, (0, self.image_height), (0, 0))
        self.rect = self.image.get_rect()
        self.x_start = 10
        self.rect.x = self.x_start
        self.rect.y = 50
        # self.rect.y = 70 + (track_idx * self.image_height)
        self.rect.y = 70

    def update(self):
        # print(self.color)
        if (self.last_bpm != self.bpm.bpm_average) or (self.last_color != self.color):
            self.render_image()
            self.dirty = 1
        self.last_bpm = self.bpm.bpm_average
        self.last_color = self.color

        elapsed_time = self.bpm.time_since_start(0.0)
        loop_length = self.bpm.get_time_for_loop()
        # print("loop_length", loop_length)
        where_line = (self.loop_width / loop_length) * elapsed_time
        self.rect.x = self.x_start + where_line

        self.dirty = 1


def main():
    if pg.get_sdl_version()[0] == 2:
        pg.mixer.pre_init(44100, 32, 2, 512)
    pg.init()
    screen = pg.display.set_mode((640, 480))

    going = True

    bpm = Bpm()
    bpm_counter = BpmCounter(bpm)
    bpm_light = BpmLight(bpm)
    bpm_line = BpmLine(bpm)
    clock = pg.time.Clock()

    pygame_dir = os.path.split(os.path.abspath(pg.__file__))[0]
    data_dir = os.path.join(pygame_dir, "examples", "data")
    sound = pg.mixer.Sound(os.path.join(data_dir, "punch.wav"))

    pg.display.set_caption("press space 4 times to adjust BPM timing")

    background = pg.Surface(screen.get_size())
    background = background.convert()
    background.fill((0, 0, 0))

    allsprites = pg.sprite.LayeredDirty([bpm_counter, bpm_light, bpm_line])
    allsprites.clear(screen, background)

    while going:
        events = pg.event.get()
        for e in events:
            print(e)
            if (
                e.type == pg.QUIT
                or e.type == pg.KEYDOWN
                and e.key in [pg.K_ESCAPE, pg.K_q]
            ):
                going = False

        bpm.update()
        # if bpm.started_beat and bpm.first_beat:
        #     sound.play()
        if bpm.started_beat:
            sound.play()

        bpm_counter.events(events)
        allsprites.update()

        rects = allsprites.draw(screen)

        # note, we don't update the display every 240 seconds.
        if rects:
            pg.display.update(rects)
        # We loop quickly so timing can be more accurate with the sounds.
        clock.tick(240)
        # print(rects)
        # print(clock.get_fps())


if __name__ == "__main__":
    main()



import random
import math
import time

import pygame as pg

from parrotjoy.tracks import Track
from parrotjoy.audiorecord import PITCHES, ONSETS, AUDIOS


SIZE = 300


class Visual(object):
    def __init__(self, x, y, color, size):
        self.x = x
        self.y = y
        self.color = color
        self.size = size

    def update(self):
        self.size -= 6
        return self



def normalize(n, range1, range2):
    delta1 = range1[1] - range1[0]
    delta2 = range2[1] - range2[0]
    return min((delta2 * (n - range1[0]) / delta1) + range2[0], range2[1])

class VideoSynth:
    def __init__(self, app):
        self.active = True

        self.screen = app.screen
        self._app = app

        self.colors = [
            (random.randint(0, 255),
            random.randint(0, 255),
            random.randint(0, 255)) for x in range(10)
        ]
        self.blend_colors = [
            pg.Color('red'),
            pg.Color('green'),
            pg.Color('blue'),
        ]
        self.visuals = []

        self.track = Track(None)

        self.size = SIZE
        self.blend = False

    def render(self):
        screen = self.screen
        screen.fill((0, 0, 0))
        new_visuals = []
        for place, visual in enumerate(self.visuals):
            if visual.size > 2:

                if self.blend:
                    surface = pg.Surface((200, 200)).convert_alpha()
                    surface.fill((0,0,0,0))
                    pg.draw.circle(surface, visual.color, (100, 100), 100, 0)
                    screen.blit(surface, (visual.x, visual.y), special_flags=pg.BLEND_ADD)
                else:
                    pg.draw.arc(screen, visual.color, [visual.x, visual.y, visual.size, visual.size], 0, math.pi/2, 2)
                    pg.draw.arc(screen, visual.color, [visual.x, visual.y, visual.size, visual.size], math.pi/2, math.pi, 2)
                    pg.draw.arc(screen, visual.color, [visual.x, visual.y, visual.size, visual.size], math.pi,3*math.pi/2, 2)
                    pg.draw.arc(screen, visual.color, [visual.x, visual.y, visual.size, visual.size], 3*math.pi/2, 2*math.pi, 2)

                new_visuals.append(visual.update())

        self.visuals = new_visuals
        return [self.screen.get_rect()]



    def update(self, elapsed_time):
        pass


    def events(self, events):
        screen_width, screen_height = self.screen.get_size()
        for event in events:
            e = event
            if event.type == PITCHES:
                pass
                if self.blend and self.blend != 2:
                    y = int(normalize(e.frequencies[0], [50, 350], [0, 2]))
                    self.visuals.append(Visual(
                        random.randint(0, screen_width),
                        random.randint(0, screen_height),
                        self.blend_colors[y],
                        size=self.size
                    ))

            elif event.type == ONSETS:
                # for onset in event.onsets:
                #     print("onset", onset)
                if self.blend == 2:
                    for onset in event.onsets:
                        self.visuals.append(Visual(
                            random.randint(0, screen_width),
                            random.randint(0, screen_height),
                            random.choice(self.blend_colors),
                            size=self.size*2
                        ))

                if not self.blend:
                    for onset in event.onsets:
                        self.visuals.append(Visual(
                            random.randint((screen_width//4), screen_width - (screen_width//2)),
                            random.randint((screen_height//4), screen_height - (screen_height//2)),
                            random.choice(self.colors),
                            size=self.size
                        ))
            elif event.type == AUDIOS:
                for audio in event.audios:
                    self.track.update(audio)


            if e.type == pg.KEYDOWN:
                if e.key == pg.K_r:
                    self.size = SIZE
                if e.key == pg.K_e:
                    self.size += -10
                if e.key == pg.K_t:
                    self.size += 10
                if e.key == pg.K_b:
                    self.blend = not self.blend
                if e.key == pg.K_n:
                    self.blend = 2




""" Some examples for generating and converting sounds for pygame.

Python 2.7, 3.6

Shows:
    - a simple 'square wave' generated
    - resampling sample rates (eg, 8363 to 44100)
    - using built in python array for making pygame.Sound samples.
    - samples at different bit sizes
    - converting from signed 8 to signed 16bit
    - how initializing the mixer changes what samples Sound needs.
    - Using the python stdlib audioop.ratecv for sample rate conversion.
    - drawing sound sample arrays as a waveform scaled into a Surface.

Square Wave
  https://en.wikipedia.org/wiki/Square_wave
MOD (file format)
  https://en.wikipedia.org/wiki/MOD_(file_format)

pygame.mixer.get_init
    https://www.pygame.org/docs/ref/mixer.html#pygame.mixer.get_init
pygame.Sound
    https://www.pygame.org/docs/ref/mixer.html#pygame.mixer.Sound

array (python stdlib)
    https://docs.python.org/3/library/array.html
wave (python stdlib)
    https://docs.python.org/3/library/wave.html
audioop.ratecv (python stdlib)
    https://docs.python.org/3/library/audioop.html?highlight=audio#audioop.ratecv

"""

from array import array
import pygame as pg

def scale_samples_to_surf(width, height, samples):
    """ Returns a generator containing (x, y) to draw a waveform.

    :param width: width of surface to scale points to.
    :param height: height of surface to scale points to.
    :param samples: an array of signed 1 byte or signed 2 byte.
    """
    # assert samples.typecode in ['h', 'b']
    # precalculate a bunch of variables, so not done in the loop.
    len_samples = len(samples)
    width_per_sample = width / len_samples
    height_1 = height - 1

    if hasattr(samples, 'typecode'):
        if samples.typecode == 'h':
            # for array typecode 'h', -32768 to 32768
            factor = 1.0 / 65532
            normalize_modifier = int(65532 / 2)
        elif samples.typecode == 'b':
            # for array typecode 'b', -127 to 127
            factor = 1.0 / 256
            normalize_modifier = int(256 / 2)
    else:
        import numpy as np
        mult = 16
        # breakpoint()
#        if samples.dtype == np.uint8
        factor = 1.0 / (65532/mult)
        normalize_modifier = int(65532 / (2 * mult))

    return ((
        int((sample_number + 1) * width_per_sample),
        int(
            (1.0 -
                (factor *
                    (samples[sample_number] + normalize_modifier)))
            * (height_1)
        ))
    for sample_number in range(len_samples))


def chunks(alist, chunk_size):
    """chunk_size chunks from alist."""
    for i in range(0, len(alist), chunk_size):
        yield alist[i:i + chunk_size]

def average_chunks(parts):
    """get the average value for each chunk"""
    for chunk in parts:
        yield sum(chunk) / len(chunk)

def resize_samples(samples, width):
    if len(samples) < width:
        return samples
    per_pixel = len(samples) / width
    chunk_size = int(round(per_pixel))
    return average_chunks(chunks(samples, chunk_size))


def draw_wave(surf,
              samples,
              wave_color = (0, 0, 0),
              background_color = pg.Color('white')):
    """Draw array of sound samples as waveform into the 'surf'.

    :param surf: Surface we want to draw the wave form onto.
    :param samples: an array of signed 1 byte or signed 2 byte.
    :param wave_color: color to draw the wave form.
    :param background_color: to fill the 'surf' with.

    waveform_surf = pg.Surface((320, 200)).convert()
    draw_wave(
        waveform_surf,
        sound.samples,
        background_color=(0,0,0),
        wave_color=(255, 255, 255))
    """
    if background_color is not None:
        surf.fill(background_color)
    width, height = surf.get_size()
    if samples.shape[0] == 0:
        return
    if 1:
        newsamples = resize_samples(samples, width)
        if hasattr(samples, 'typecode'):
            assert samples.typecode in ['h', 'b']
            samples = array(samples.typecode, newsamples)
        else:
            samples = list(newsamples)

    points = tuple(scale_samples_to_surf(width, height, samples))
    if points:
        pg.draw.lines(surf, wave_color, False, points)






"""Recording audio and doing some analysis on it as we go.
"""

from queue import Queue

import numpy as np
import aubio
import pygame as pg
from pygame.midi import (
    midi_to_frequency,
    midi_to_ansi_note,
)

pygame = pg

from pygame._sdl2 import (
    AUDIO_F32,
    get_audio_device_names,
    AudioDevice,
    AUDIO_ALLOW_FORMAT_CHANGE,
)

AUDIOS = pg.event.custom_type()
ONSETS = pg.event.custom_type()
PITCHES = pg.event.custom_type()


class AudioRecord:
    """Recording audio and doing some analysis on it as we go.

    It looks for onsets (beats), pitch detection (notes, and midi frequencies)
    """

    def __init__(self, inputdevice):
        self.inputdevice = inputdevice
        self.audio_going = True
        self.onset_queue = Queue()
        self.audio_queue = Queue()
        self.pitch_queue = Queue()

        self.callback = lambda audiodevice, stream: self.on_data(audiodevice, stream)

    def update(self):
        """To be called in a pygame event loop."""
        audios = []
        while not self.audio_queue.empty():
            # print('-----audio from queue')
            # print(time.time() - time_start)
            audio = self.audio_queue.get()
            audios.append(audio)

        if audios:
            pg.event.post(pg.event.Event(AUDIOS, audios=audios))

        onsets = []
        while not self.onset_queue.empty():
            onsets.append(self.onset_queue.get())

        if onsets:
            pg.event.post(pg.event.Event(ONSETS, onsets=onsets))

        pitches = []
        notes = []
        frequencies = []
        while not self.pitch_queue.empty():

            # print("pitch, confidence", pitch, confidence, midi_to_ansi_note(pitch))
            pitch = self.pitch_queue.get()
            pitches.append(pitch)
            notes.append(midi_to_ansi_note(pitch))
            frequencies.append(midi_to_frequency(pitch))

        if pitches:
            pg.event.post(
                pg.event.Event(
                    PITCHES, pitches=pitches, notes=notes, frequencies=frequencies
                )
            )

    def on_data(self, audiodevice, audiobuffer):
        self.audio_queue.put(bytes(audiobuffer))
        # print(type(audiobuffer), len(audiobuffer))
        signal = np.frombuffer(audiobuffer, dtype=np.float32)
        # print(audiodevice)
        onset = self.onset(signal)
        if onset > 1:
            # print("%f" % self.onset.get_last_s())
            self.onset_queue.put(True)

        pitch = self.pitch_o(signal)[0]
        # pitch = int(round(pitch))
        if int(round(pitch)):
            confidence = self.pitch_o.get_confidence()
            # print("pitch, confidence", pitch, confidence, midi_to_ansi_note(pitch))
            self.pitch_queue.put(pitch)

    def start(self):
        FORMAT = AUDIO_F32
        CHANNELS = 1
        RATE = 44100
        CHUNK = 512

        just_names = get_audio_device_names(iscapture=1)
        print(just_names)
        devicename = (
            self.inputdevice if self.inputdevice in just_names else just_names[0]
        )
        print(f"Using: {devicename}")

        # onset detection.
        win_s = 1024
        hop_s = win_s // 2
        self.onset = aubio.onset("default", win_s, hop_s, RATE)

        tolerance = 0.8
        self.pitch_o = aubio.pitch("yin", win_s, hop_s, RATE)
        self.pitch_o.set_unit("midi")
        self.pitch_o.set_tolerance(tolerance)

        self.audio_device = AudioDevice(
            devicename=devicename,
            iscapture=1,
            frequency=RATE,
            audioformat=AUDIO_F32,
            numchannels=CHANNELS,
            chunksize=CHUNK,
            allowed_changes=AUDIO_ALLOW_FORMAT_CHANGE,
            callback=self.callback,
        )

        self.audio_device.pause(0)

    def __del__(self):
        self.audio_device.pause(1)
        self.audio_device.close()



import random, math, os, glob, string, time, pygame
from pygame.locals import *
pg = pygame
from parrotjoy.audiorecord import PITCHES, ONSETS, AUDIOS

#constants
WINSIZE = [1920, 1080]
WINSIZE = [2048, 1280]
WINSIZE = [640, 480]
WINSIZE = [2448/2, 2448/2]
WINSIZE = [1000, 1000]
WINCENTER = [320, 240]

WINSIZE = [1024, 768]
WINCENTER = [int(1024 / 2), int(768 / 2)]
NUMSTARS = 150



def load_letters(extra_sizes = [(100, 100), (768, 768)]):
    letter_path = os.path.join('data', 'images', 'letters')
    # fnames = glob.glob(os.path.join(letter_path, '*.png'))
    # image_paths = {os.path.split(f)[1].split('.png')[0]: f for f in fnames}
    fnames = glob.glob(os.path.join(letter_path, '*.jpg'))
    image_paths = {os.path.split(f)[1].split('.jpg')[0]: f for f in fnames}

    images = {}

    def load_image(letter):
        surf = pygame.image.load(image_paths[letter])
        images[letter] = surf.convert()
        for x, y in extra_sizes:
            images[letter + '%sx%s' % (x, y)] = pygame.transform.smoothscale(surf, (x, y)).convert()

    # def load_image(letter):
    #     surf = pygame.image.load(image_paths[letter])
    #     if surf.get_size() != (2448, 2448):
    #         surf = pygame.transform.smoothscale(surf, (2448, 2448))
    #     images[letter] = surf.convert()
    #     for x, y in extra_sizes:
    #         images[letter + '%sx%s' % (x, y)] = pygame.transform.smoothscale(surf, (x, y)).convert()

    pygame.threads.tmap(load_image, image_paths)
    return images



def init_star():
    "creates new star values"
    dir = random.randrange(100000)
    velmult = random.random()*.6+.4
    vel = [math.sin(dir) * velmult, math.cos(dir) * velmult]
    return vel, WINCENTER[:]


def initialize_stars():
    "creates a new starfield"
    stars = []
    for x in range(NUMSTARS):
        star = init_star()
        vel, pos = star
        steps = random.randint(0, WINCENTER[0])
        pos[0] = pos[0] + (vel[0] * steps)
        pos[1] = pos[1] + (vel[1] * steps)
        vel[0] = vel[0] * (steps * .09)
        vel[1] = vel[1] * (steps * .09)
        stars.append(star)
    move_stars(stars)
    return stars


def draw_stars(surface, stars, color):
    "used to draw (and clear) the stars"
    for vel, pos in stars:
        pos = (int(pos[0]), int(pos[1]))
        surface.set_at(pos, color)


def move_stars(stars):
    "animate the star values"
    for vel, pos in stars:
        pos[0] = pos[0] + vel[0]
        pos[1] = pos[1] + vel[1]
        if not 0 <= pos[0] <= WINSIZE[0] or not 0 <= pos[1] <= WINSIZE[1]:
            vel[:], pos[:] = init_star()
        else:
            vel[0] = vel[0] * 1.05
            vel[1] = vel[1] * 1.05

def normalize(n, range1, range2):
    delta1 = range1[1] - range1[0]
    delta2 = range2[1] - range2[0]
    return min((delta2 * (n - range1[0]) / delta1) + range2[0], range2[1])



PAUSE_TIME = 0.2

class Strawberries:

    def __init__(self, app):
        self.active = True

        self.screen = app.screen
        self._app = app


        random.seed()
        self.stars = initialize_stars()


        self.letters = load_letters()
        self.letter_i = 0
        self.ascii_lowercase = string.ascii_lowercase

        self.pause_on_letter = time.time()
        self.random_pause = False

        self.pause_time = PAUSE_TIME



    def events(self, events):
        for e in events:
            if e.type == MOUSEBUTTONDOWN and e.button == 1:
                WINCENTER[:] = list(e.pos)
            if e.type == PITCHES:
                y = normalize(e.frequencies[0], [50, 350], [0, 768])
                WINCENTER[:] = [int(1024 / 2), y]
                # print(y, e.frequencies[0])
            if e.type == ONSETS:
                if not self.random_pause:
                    self.pause_on_letter = time.time()

            if e.type == pg.KEYDOWN:
                if e.key == pg.K_r:
                    self.random_pause = not self.random_pause
                    self.pause_time = PAUSE_TIME
                if e.key == pg.K_e:
                    self.pause_time += -0.02
                if e.key == pg.K_t:
                    self.pause_time += 0.02



    def update(self, elapsed_time):
        """ return True to let other scenes update. False to only us update.
        """
        pass

    def render(self):
        """ return rects.

            If scene.propagate_render is True, the render will
                continue to be propagated.
        """
        screen = self.screen

        white = 255, 240, 200
        # black = 20, 20, 40
        black = 0, 0, 0
        screen.fill(black)

        x = int((1024 - 768) // 2)
        key = self.ascii_lowercase[self.letter_i] + '768x768'
        if not key in self.letters:
            return []
        screen.blit(self.letters[key], (x, 0))



        # print (pause_on_letter)
        if self.pause_on_letter:
            # print('yoooooop', pause_on_letter + 1.0, time.time())
            if self.pause_on_letter + self.pause_time < time.time():
                # print('finished')
                self.pause_on_letter = 0
        else:
            self.letter_i += 1
            if self.letter_i >= len(self.ascii_lowercase):
                self.letter_i = 0
            if self.random_pause:
                if random.randint(0, 10) == 5:
                    self.pause_on_letter = time.time()

        draw_stars(screen, self.stars, black)
        move_stars(self.stars)
        draw_stars(screen, self.stars, white)
        return [self.screen.get_rect()]



""" Command line interface.
"""
import sys, os

#these directories will be used if parrotjoy cannot find
#the directories in the location of the main program
if sys.platform == "win32":
    DATADIR = "C:\\Program Files\\parrotjoy"
    CODEDIR = "C:\\Program Files\\parrotjoy\\code"
else:
    DATADIR = "/usr/share/games/parrotjoy"
    CODEDIR = "/usr/lib/games/parrotjoy"



def main():
    #figure out our directories
    global DATADIR, CODEDIR

    # first try to get the path from the parrotjoy package.
    try:
        import parrotjoy
        if type(parrotjoy.__path__) == list:
            localpath = os.path.abspath(parrotjoy.__path__[0])
        else:
            localpath = os.path.abspath(parrotjoy.__path__._path[0])
    except ImportError:
        localpath = os.path.split(os.path.abspath(sys.argv[0]))[0]

    testdata = localpath
    testcode = os.path.join(localpath, '.')

    if os.path.isdir(os.path.join(testdata, 'data')):
        DATADIR = testdata
    if os.path.isdir(testcode):
        CODEDIR = testcode

    # pyinstaller uses this variable to store the path
    #   where it extracts data to.
    pyinstaller_path = getattr(sys, '_MEIPASS', None)
    if pyinstaller_path:
        DATADIR = os.path.join(pyinstaller_path, 'data')
    else:
        #apply our directories and test environment
        os.chdir(DATADIR)
        sys.path.insert(0, CODEDIR)
        checkdependencies()

    #run game and protect from exceptions
    try:
        # import pdb;pdb.set_trace()
        try:
            from .main import main
        except ImportError:
            from parrotjoy.main import main
        import pygame as pg
        main(sys.argv)
    except KeyboardInterrupt:
        print('Keyboard Interrupt (Control-C)...')
    except:
        #must wait on any threading
        # if game.thread:
        #     game.threadstop = 1
        #     while game.thread:
        #         pg.time.wait(10)
        #         print('waiting on thread...')

        # we need to enable a debug handler for release.
        # exception_handler()
        # if game.DEBUG:
        #     raise
        raise



def checkdependencies():
    "only returns if everything looks ok"
    msgs = []

    #make sure this looks like the right directory
    if not os.path.isdir(CODEDIR):
        msgs.append('Cannot locate parrotjoy modules')
    if not os.path.isdir('data'):
        msgs.append('Cannot locate parrotjoy data files')

    #first, we need python >= 2.7
    if sys.hexversion < 0x2070000:
        errorbox('Requires Python-2.7 or Greater')

    #is correct pg found?
    try:
        import pygame as pg
        if pg.ver < '1.9.4':
            msgs.append('Requires pygame 1.9.4 or Greater, You Have ' + pg.ver)
    except ImportError:
        msgs.append("Cannot import pygame, install version 1.9.4 or higher")
        pg = None

    #check that we have FONT and IMAGE
    if pg:
        if not pg.font:
            msgs.append('pg requires the SDL_ttf library, not available')
        if not pg.image or not pg.image.get_extended():
            msgs.append('pg requires the SDL_image library, not available')

    if msgs:
        msg = '\n'.join(msgs)
        errorbox(msg)



#Pretty Error Handling Code...

def __windowsbox(title, message):
    raise ImportError #the MessageBox command is crashing!
    import win32ui, win32con
    win32ui.MessageBox(message, title, win32con.MB_ICONERROR)

def __wxpythonbox(title, message):
    import wxPython.wx as wx
    class LameApp(wx.wxApp):
        def OnInit(self): return 1
    app = LameApp()
    dlg = wx.wxMessageDialog(None, message, title, wx.wxOK|wx.wxICON_EXCLAMATION)
    dlg.ShowModal()
    dlg.Destroy()

def __tkinterbox(title, message):
    import tkinter, tkinter.messagebox
    tkinter.Tk().wm_withdraw()
    tkinter.messagebox.showerror(title, message)


def __pgbox(title, message):
    try:
        import pygame as pg
        pg.quit() #clean out anything running
        pg.display.init()
        pg.font.init()
        screen = pg.display.set_mode((460, 140))
        pg.display.set_caption(title)
        font = pg.font.Font(None, 18)
        foreg, backg, liteg = (0, 0, 0), (180, 180, 180), (210, 210, 210)
        ok = font.render('Ok', 1, foreg, liteg)
        okbox = ok.get_rect().inflate(200, 10)
        okbox.centerx = screen.get_rect().centerx
        okbox.bottom = screen.get_rect().bottom - 10
        screen.fill(backg)
        screen.fill(liteg, okbox)
        screen.blit(ok, okbox.inflate(-200, -10))
        pos = [10, 10]
        for text in message.split('\n'):
            if text:
                msg = font.render(text, 1, foreg, backg)
                screen.blit(msg, pos)
            pos[1] += font.get_height()
        pg.display.flip()
        stopkeys = pg.K_ESCAPE, pg.K_SPACE, pg.K_RETURN, pg.K_KP_ENTER
        while 1:
            e = pg.event.wait()
            if e.type == pg.QUIT or \
                       (e.type == pg.KEYDOWN and e.key in stopkeys) or \
                       (e.type == pg.MOUSEBUTTONDOWN and okbox.collidepoint(e.pos)):
                break
        pg.quit()
    except pg.error:
        raise ImportError

handlers = __pgbox, __tkinterbox, __wxpythonbox, __windowsbox

def __showerrorbox(message):
    title = os.path.splitext(os.path.split(sys.argv[0])[1])[0]
    title = title.capitalize() + ' Error'
    for e in handlers:
        try:
            e(title, message)
            break
        except (ImportError, NameError, AttributeError):
            pass

def errorbox(message):
    message = str(message)
    if not message: message = 'Error'
    __showerrorbox(message)
    sys.stderr.write('ERROR: ' + message + '\n')
    raise SystemExit

def exception_handler():
    import traceback
    type, info, trace = sys.exc_info()
    tracetop = traceback.extract_tb(trace)[-1]
    tracetext = 'File %s, Line %d' % tracetop[:2]
    if tracetop[2] != '?':
        tracetext += ', Function %s' % tracetop[2]
    exception_message = '%s:\n%s\n\n%s\n"%s"'
    message = exception_message % (str(type), str(info), tracetext, tracetop[3])
    if type not in (KeyboardInterrupt, SystemExit):
        __showerrorbox(message)
    raise



if __name__ == '__main__':
    main()



"""Basically doing sound effects, graphics and music-ing.

Does paths, loading, caching, converting and such... for us.

gfx('parrot.png', convert=True) will look into the data/images/ folder, load the image,
and maybe even convert the image for you.

Next place you use gfx('parrot.png', convert=True) will give you the same cached image.

This means when using gfx, you just use it. 
Don't not worry about paths, caching, converting and such things.
"""
import os
import pygame as pg

# Our secret stashes.
_sfx_cache = {}
_gfx_cache = {}


def data_path() -> str:
    """Returns file system path for where the data at."""
    if os.path.exists("data"):
        path = "data"
    else:
        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "data",
        )
    return path


def music(
    music: str | None = None,
    load: bool = True,
    play: bool = True,
    stop: bool | None = False,
) -> None:
    """Basically music-ing. music('song.mp3')

    :param music: the music file to play. Relative to data/sounds.
    :param load: Start loading the music. If music is loaded already it's faster to play.
    :param play: Play music.
    :param stop: Stop music.

    #### Example

    ```python
    music('bla.ogg')
    music(stop=True)
    ```
    """
    # perhaps the mixer is not included or initialised.
    if pg.mixer and pg.mixer.get_init():
        if load and not stop:
            pg.mixer.music.load(music_path(music))
        if play and stop is None or stop is False:
            pg.mixer.music.play()
        elif stop:
            pg.mixer.music.stop()


def music_path(music: str) -> str:
    """Returns file system path for where the given music is at."""
    path = os.path.join(data_path(), "sounds", music)
    return path


def gfx(image: str, convert: bool = False, convert_alpha: bool = False) -> pg.Surface:
    """Basically graphics using. Loading/converting/cached. gfx('parrot.png')

    :param image: looks in data/images/ for the image to load.
    :param convert: convert to the most meow
    :param convert_alpha: convert_alpha()
    :param play: Play music.
    :param stop: Stop music.

    #### Example

    ```python
    gfx('parrot.png')
    gfx('parrot.png', convert=True)
    gfx('parrot.png', convert_alpha=True)
    ```
    """
    global _gfx_cache
    gfx_key = (image, convert, convert_alpha)
    if gfx_key in _gfx_cache:
        return _gfx_cache[gfx_key]

    path = os.path.join(data_path(), "images", image)
    asurf = pg.image.load(path)
    if convert:
        asurf = asurf.convert()
    if convert_alpha:
        asurf = asurf.convert_alpha()

    _gfx_cache[gfx_key] = asurf
    return asurf


def sfx(snd: str, play: bool = False, stop: bool = False) -> pg.mixer.Sound:
    """Basically sound using. Loading/converting/cached. sfx('parrot.mp3')

    :param snd: looks in data/sounds/ for the sound to load.
    :param play: Play sound.
    :param stop: Stop sound.

    #### Example

    ```python
    sfx('parrot.mp3')
    sfx('parrot.png', play=True)
    sfx('parrot.png', convert_alpha=True)
    ```
    """

    global _sfx_cache
    snd_key = snd
    if snd_key in _sfx_cache:
        asound = _sfx_cache[snd_key]
    else:
        path = os.path.join(data_path(), "sounds", snd)
        asound = pg.mixer.Sound(path)
        _sfx_cache[snd_key] = asound

    # print(snd_key, play, stop, time.time())
    if play:
        asound.play()
    if stop:
        asound.stop()
    return asound



import sys
import io
import random
import math
import time

import numpy as np
import pygame as pg
from pygame.midi import frequency_to_midi, midi_to_frequency, midi_to_ansi_note

from tracks import Track
from audiorecord import AudioRecord, PITCHES, ONSETS, AUDIOS
from scenes.videosynth import VideoSynth
from scenes.looper.looper import Looper
from scenes.strawberries import Strawberries


DEVICENAME_INPUT = None
DEVICENAME_OUTPUT = None
# DEVICENAME_INPUT = b'Scarlett 2i4 USB'
# DEVICENAME_OUTPUT = b'Scarlett 2i4 USB'


class App:
    FLAGS = pg.FULLSCREEN
    FLAGS = 0
    FLAGS = pg.SCALED | pg.FULLSCREEN
    WIDTH = 1024
    HEIGHT = 768
    # FPS = 30
    FPS = 240

    def __init__(self):

        # pg.mixer.pre_init(44100, 32, 2, 1024, devicename=DEVICENAME_OUTPUT)
        pg.mixer.pre_init(
            44100, 32, 2, 1024, devicename=DEVICENAME_OUTPUT, allowedchanges=0
        )
        pg.init()
        pg.font.init()

        self.audio_thread = AudioRecord(inputdevice=DEVICENAME_INPUT)
        self.audio_thread.start()

        joys = {}
        for x in range(pg.joystick.get_count()):
            j = pg.joystick.Joystick(x)
            joys[x] = j
            j.init()

        self.screen = pg.display.set_mode((self.WIDTH, self.HEIGHT), self.FLAGS)
        self.clock = pg.time.Clock()

        self.running = True

        self.videosynth = VideoSynth(self)
        self.videosynth.active = False
        self.looper = Looper(self)
        self.strawberries = Strawberries(self)
        self.strawberries.active = False
        self.scenes = [self.looper, self.videosynth, self.strawberries]

        self.gifmaker = None

    def looper_active(self):
        self.looper.active = True
        self.looper.redraw()
        self.videosynth.active = False
        self.strawberries.active = False
        self.FPS = 240

    def videosynth_active(self):
        self.looper.active = False
        self.videosynth.active = True
        self.strawberries.active = False
        self.FPS = 30

    def strawberries_active(self):
        self.looper.active = False
        self.videosynth.active = False
        self.strawberries.active = True
        self.FPS = 30

    def events(self, events):
        """ """
        for event in events:
            if (
                event.type == pg.QUIT
                or event.type == pg.KEYDOWN
                and event.key in [pg.K_ESCAPE, pg.K_q]
            ):
                self.running = False
            if event.type == pg.KEYDOWN and event.key == pg.K_0:
                self.looper_active()
            elif event.type == pg.KEYDOWN and event.key == pg.K_9:
                self.videosynth_active()
            elif event.type == pg.KEYDOWN and event.key == pg.K_8:
                self.strawberries_active()

        if not self.running:
            return
        # print(self.scenes[::-1], events)
        for scene in self.scenes[::-1]:
            if scene.active:
                scene.events(events)

    def update(self, elapsed_time):
        """update all the scenes until one returns False."""
        self.audio_thread.update()

        for scene in self.scenes[::-1]:
            if scene.active:
                if not scene.update(elapsed_time):
                    break

        self.clock.tick(self.FPS)

    def render(self):
        """Render the highest active scene.

        If scene.propagate_render is True, the render will
            continue to be propagated.
        """

        all_rects = []
        for scene in self.scenes[::-1]:
            if scene.active:
                rects = scene.render()
                if rects is not None:
                    all_rects.extend(rects)
                if not getattr(scene, "propagate_render", False):
                    break
        # print(all_rects)
        pg.display.update(all_rects)

    def main(self):
        """ """

        elapsed_time = 0
        while self.running:
            fs = time.time()
            self.events(pg.event.get())
            self.update(elapsed_time)
            self.render()
            elapsed_time = (time.time() - fs) * 1000
            if self.gifmaker is not None:
                self.gifmaker.update(events, self.screen)

        self.audio_thread.audio_going = False

        pg.quit()


def main(args):
    try:
        gamemain(args)
    except KeyboardInterrupt:
        print("Keyboard Interrupt...")
        print("Exiting")


def gamemain(args):

    app = App()
    app.main()


if __name__ == "__main__":
    main(sys.argv)



class Scene:
    def __init__(self, app):
        self.active = True

        self.screen = app.screen
        self._app = app

    def events(self, events):
        pass

    def update(self, elapsed_time):
        """ return True to let other scenes update. False to only us update.
        """
        pass

    def render(self):
        """ return rects.

            If scene.propagate_render is True, the render will
                continue to be propagated.
        """
        pass


from setuptools import setup, find_packages
import os
from os import path
import sys

here = path.abspath(path.dirname(__file__))

name = "parrotjoy"

from io import open
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

freeze_cmds = ["bdist_dmg", "bdist_msi", 'build_exe', 'bdist_mac']
if any(x in sys.argv for x in freeze_cmds):
    # https://cx-freeze.readthedocs.io/en/latest/distutils.html
    # Note: we needed to use git cx_freeze because it does not work on mac py3.7.
    #     git+https://github.com/anthony-tuininga/cx_Freeze.git
    #
    from cx_Freeze import setup, Executable

    # Dependencies are automatically detected, but it might need fine tuning.
    build_exe_options = {
        "packages": [
            "os", "pygame", "sys", "random"
        ],
        "excludes": ["tkinter"],
    }
    # GUI applications require a different base on Windows (the default is for a
    # console application).
    base = None
    if sys.platform.startswith("win"):
        base = "Win32GUI"

    options = {
        "build_exe": build_exe_options
    }
    executables = [Executable("run_game.py", base=base)]
    print("options, executables, base", options, executables, base)
else:
    options = {}
    executables = []




setup(
    name=name,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries :: pygame',
        'Topic :: Games/Entertainment :: Arcade',
    ],
    license='LGPL',
    author='parrotjoy team',
    author_email='parrotjoy@pygame.org',
    maintainer='parrotjoy team',
    maintainer_email='parrotjoy@pygame.org',
    description='parrotjoy',
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
    options=options,
    executables=executables,
    package_dir={'parrotjoy': 'parrotjoy'},
    packages=find_packages(),
    # package_data={'parrotjoy': []},
    url='https://github.com/pygame/parrotjoy',
    install_requires=[
        "pygame",
    ],
    version='0.0.1.dev0',
    entry_points={
        'console_scripts': [
            'parrotjoy=parrotjoy.cli:main',
        ],
    },
)



import sys
import io
import random
import math
from threading import Thread
from queue import Queue
import time


import numpy as np
import aubio
import pygame as pg
pygame = pg



def make_pygame_sound(abuffer=None, array=None):
    # convert the data from float32 into int32.
    # float32_data = np.frombuffer(b''.join(self.frames), dtype=np.float32)
    if array is None:
        float32_data = np.frombuffer(abuffer, dtype=np.float32)
    else:
        float32_data = array
    stereo = np.vstack([float32_data, float32_data]).T
    #print(stereo)
    reshaped_float32_data = float32_data.reshape(-1, 1)
    stereo = np.hstack((reshaped_float32_data, reshaped_float32_data))
    # print (stereo.shape)
    # print (pg.mixer.get_init())
    asound = pg.mixer.Sound(array=stereo)
    return asound


def mix_audio(audio1, audio2):
    """ mix two pieces of audio together.

    The output should be as long as the final one.
    """

    audio_is_longer = len(audio1) >= len(audio2)
    longer_audio = audio1 if audio_is_longer else audio2
    shorter_audio = audio2 if audio_is_longer else audio1

    common_part = longer_audio[:len(shorter_audio)]
    left_overs = longer_audio[len(shorter_audio):]

    # mixed_audio = (common_part * 0.5 + shorter_audio * 0.5)
    mixed_audio = (common_part + shorter_audio)

    newaudio = np.concatenate((mixed_audio, left_overs))
    return newaudio


def trim_audio(audio, start_trim=0, end_trim=0, rate=44100, channels=1):
    """ return the audio with the start_trim, and end_trim removed.

    :param audio: a numpy array of samples.
    :param rate: samples per second.
    :param start_trim: seconds from the start of the track to use.
    :param end_trim: seconds from the end of the track to use.
    """
    if end_trim > 0:
        return audio[int(rate * start_trim):-int(rate * end_trim)]
    else:
        return audio[int(rate * start_trim):]




class Track:

    def __init__(self, bpm):
        self.bpm = bpm
        self.add_to_mode = False
        self.frames = []
        self.audios = []
        self.audios_untrimmed = []

        self.sounds = []
        self.add_to_mode = False
        self.recording = False
        """
        recording
            - False, not recording
            - 1,     starting a new track
            - 2,     adding to
        """

        self.start_trim = 0
        """ seconds from the start of the track to use.
        """
        self.end_trim = 0
        """ seconds from the end of the track to use.
        """

        self.erased = False
        self.finished = False
        self.trimmed = False

        self._start_new_next = False
        self._add_to_next = False

    def start_new_next(self):
        print("track: start_new_next")
        self._start_new_next = True
        self.recording = 1
    def add_to_next(self):
        print("track: add_to_next")
        self._add_to_next = True
        self.recording = 1

    def start_new(self):
        print("track: start_new")
        self.frames = []
        self.audios = []
        self.audios_untrimmed = []
        self.sounds = []
        self.add_to_mode = False
        self.recording = 2

    def add_to(self):
        print("track: add_to")
        self.frames_add_to = []
        self.add_to_mode = True
        self.recording = 2

    def handle_start_new_next(self):
        # handle add_to_next() and start_new_next() calls.
        # We start them on the first beat of the loop.
        # print("handle_start_new_next", self.bpm.first_beat, self.bpm.started_beat)
        if self.bpm is not None:
            if self.bpm.first_beat and self.bpm.started_beat:
                if self._start_new_next:
                    self.start_new()
                    self._start_new_next = False
                if self._add_to_next:
                    self.add_to()
                    self._add_to_next = False

    def erase(self):
        print("track: erase")
        self.frames = []
        self.audios = []
        self.audios_untrimmed = []
        self.sounds = []
        self.add_to_mode = False
        self.recording = False
        self.erased = True
        self.trimmed = False

    def trim_audio(self, start=0, end=0):
        self.start_trim += start
        self.end_trim += end
        if self.end_trim - 0.000001 < 0:
            self.end_trim = 0
        if self.end_trim > 2:
            # TODO: if end_trim > sound_length
            self.end_trim = 2


        if self.start_trim - 0.000001 < 0:
            self.start_trim = 0
        if self.start_trim > 2:
            # TODO: if start_trim > sound_length
            self.start_trim = 2


        print("trim_audio", start, end, self.start_trim, self.end_trim)
        if len(self.audios_untrimmed):
            frames = self.audios_untrimmed[0]
        else:
            frames = self.audios[0]

        audio = np.frombuffer(frames, dtype=np.float32)
        trimmed_audio = trim_audio(audio, self.start_trim, self.end_trim)
        self.sounds = [
            make_pygame_sound(array=trimmed_audio)
        ]
        if not len(self.audios_untrimmed):
            self.audios_untrimmed = self.audios[:]

        self.audios = [trimmed_audio.tobytes()]
        self.trimmed = True

    def finish(self):
        print("track: finish")
        self.recording = False
        # print(f'duration_seconds:{audio.duration_seconds}')
        if self.add_to_mode:
            all_frames = b''.join(self.frames_add_to)
        else:
            all_frames = b''.join(self.frames)

        self.audios.append(all_frames)

        if self.add_to_mode:
            audio1 = np.frombuffer(self.audios[0], dtype=np.float32)
            if len(self.audios) == 2:
                audio2 = np.frombuffer(self.audios[1], dtype=np.float32)
                mixed_audio = mix_audio(audio1, audio2)
            else:
                mixed_audio = mix_audio(audio1, audio1)
            self.audios = [mixed_audio.tobytes()]
            self.sounds = [
                make_pygame_sound(array=mixed_audio)
            ]
        else:
            self.sounds = [make_pygame_sound(all_frames)]
        self.add_to_mode = False
        self.finished = True


    def play(self):
        for sound in self.sounds:
            sound.play()

    def stop(self):
        for sound in self.sounds:
            sound.stop()

    def update(self, audiobuffer):
        # print(f'id: {id(audiobuffer)}')
        if self.recording:
            if self.recording == 2 and audiobuffer is not None:
                if self.add_to_mode:
                    self.frames_add_to.append(audiobuffer)
                else:
                    self.frames.append(audiobuffer)
        if self.erased:
            self.erased = False
        if self.trimmed:
            self.trimmed = False
        if self.finished:
            print("update: finished")
            self.finished = False

        self.handle_start_new_next()





"""For grabbing camera data in a thread and putting pygame Surfaces on a queue.

https://docs.opencv.org/2.4/modules/highgui/doc/reading_and_writing_images_and_video.html#videocapture

TODO:

- A window per camera for the demo might be good.
- Add a --what cmdline argument (or a better name) for selecting a camera by index.
"""
import pygame as pg
import pygame.camera

pg.camera = pygame.camera  # hack for pylance thinking pygame.camera is unused.
# from pygame.locals import *
import numpy as np
import sys
from threading import Thread
from queue import Queue

# If we can use cv2 use that by default.
USE_CV2 = False

try:
    import cv2

    USE_CV2 = True
except:
    pass


class VideoThreadCV2(Thread):
    """Camera Surfaces into queue."""

    def __init__(
        self,
        what: int | str = 1,  
        width: int = 640,  # 
        height: int = 480,  # 
    ):
        """Camera Surfaces into a queue.
        
        :param int | str what: What to open. Either an index or a string path.
        :param int width: Width in pixels of Surface to put in the Queue.
        :param int height: Height in pixels of Surface to put in the Queue.
        """
        Thread.__init__(self)
        self.width = width
        self.height = height

        self.going = True
        self.queue = Queue()
        self.camera = cv2.VideoCapture(what)
        if not self.camera.isOpened():
            self.going = False
        self.camera.set(cv2.CAP_PROP_FRAME_WIDTH, width)
        self.camera.set(cv2.CAP_PROP_FRAME_HEIGHT, height)

    def stop(self):
        """Stop going."""
        self.going = False

    def run(self):
        """While going put video frames on the queue."""
        while self.going:
            try:
                ret, frame = self.camera.read()
                if ret:
                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    frame = cv2.rotate(frame, cv2.ROTATE_90_CLOCKWISE)
                    frame = cv2.flip(frame, 1)
                    surface = pg.surfarray.make_surface(frame)
                    self.queue.put(surface)
                else:
                    self.queue.put(False)
            except KeyboardInterrupt:
                break
        self.camera = None


class VideoThreadPygame(Thread):
    """Camera Surface into queue."""

    def __init__(self, what=0, width=640, height=480):
        """Camera Surfaces into a queue.
        
        :param int | str what: What to open. Either an index or a string path.
        :param int width: Width in pixels of Surface to put in the Queue.
        :param int height: Height in pixels of Surface to put in the Queue.
        """
        Thread.__init__(self)
        self.width = width
        self.height = height
        self.size = (width, height)

        self.going = True
        self.queue = Queue()

        try:
            pg.camera.init()
        except:
            self.going = False
            return

        cams = pg.camera.list_cameras()
        if cams:
            self.camera = pg.camera.Camera(cams[what], self.size, "RGB")
            self.going = True
        else:
            self.camera = None
            self.going = False

        if self.camera is not None:
            self.camera.start()

        self.snapshot = pg.surface.Surface(self.size, 0, pg.display.get_surface())

    def stop(self):
        """Stop going."""
        self.going = False

    def run(self):
        """While going put video frames on the queue."""
        while self.going:
            try:
                # self.snapshot = self.camera.get_image(self.snapshot)
                surface = self.camera.get_image()
                if surface is not None:
                    self.queue.put(surface)
            except KeyboardInterrupt:
                break
        self.camera = None


if USE_CV2:
    VideoThread = VideoThreadCV2
else:
    VideoThread = VideoThreadPygame


def main():

    import argparse

    def init_argparse() -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
            usage="%(prog)s [OPTION]",
            description="View a camera. Optionally choose --pygame or --cv2",
        )
        parser.add_argument(
            "--pygame",
            dest="pygame",
            action="store_const",
            const=True,
            default=not USE_CV2,
            help="use pygame camera",
        )
        parser.add_argument(
            "--cv2",
            dest="cv2",
            action="store_const",
            const=True,
            default=USE_CV2,
            help="use cv2 camera",
        )

        return parser

    parser = init_argparse()
    args = parser.parse_args()

    pg.init()
    screen = pg.display.set_mode([640, 480])
    # video_thread = VideoThread('some_movie.mp4')
    if args.pygame:
        print("using pygame camera")
        video_thread = VideoThreadPygame(0, 640, 480)
    else:
        print("using cv2 camera")
        video_thread = VideoThreadCV2(1, 640, 480)

    video_thread.daemon = True
    video_thread.start()
    clock = pg.time.Clock()
    try:
        while True:

            for event in pg.event.get():
                if event.type == pg.KEYDOWN:
                    sys.exit(0)

            while not video_thread.queue.empty():
                surface = video_thread.queue.get()
                if surface:
                    screen.blit(surface, (0, 0))
            pg.display.flip()
            # clock.tick(30)

    except (KeyboardInterrupt, SystemExit):
        pg.quit()

    cv2.destroyAllWindows()


if __name__ == "__main__":
    main()
